@@ -66,12 +66,19 @@ extern struct simplelock afs_rxglobal_lock;
 #define VN_RELE(vp)		vrele(vp)
 #define VN_HOLD(vp)		VREF(vp)
 
+#ifdef AFS_FBSD60_ENV
+#undef IsAfsVnode
+#define IsAfsVnode(v) ((v)->v_op == &afs_vnodeops)
+extern struct vop_vector afs_vnodeops;
+#endif
+
 #if defined(AFS_FBSD50_ENV)
 #define VT_AFS		"afs"
 #define VROOT		VV_ROOT
 #define v_flag		v_vflag
 #define osi_curcred()	(curthread->td_ucred)
-#define afs_suser()	(!suser(curthread))
+#define afs_suser(x)	(!suser(curthread))
+#define getpid()	(curthread->td_proc->p_pid)
 #define simple_lock(x)	mtx_lock(x)
 #define simple_unlock(x) mtx_unlock(x)
 #define        gop_rdwr(rw,gp,base,len,offset,segflg,unit,cred,aresid) \
@@ -85,8 +92,8 @@ extern struct mtx afs_global_mtx;
 extern struct lock afs_global_lock;
 
 #define osi_curcred()	(curproc->p_cred->pc_ucred)
-#define afs_suser()	(!suser(curproc))
-#define getpid()	curproc
+#define afs_suser(x)	(!suser(curproc))
+#define getpid()	(curproc->p_pid)
 #define        gop_rdwr(rw,gp,base,len,offset,segflg,unit,cred,aresid) \
   vn_rdwr((rw),(gp),(base),(len),(offset),(segflg),(unit),(cred),(aresid), curproc)
 extern struct proc *afs_global_owner;
@@ -107,10 +114,6 @@ extern struct proc *afs_global_owner;
 #define ISAFS_GLOCK() (afs_global_owner == curproc && curproc)
 #endif /* FBSD50 */
 
-#define AFS_RXGLOCK()
-#define AFS_RXGUNLOCK()
-#define ISAFS_RXGLOCK() 1
-
 #undef SPLVAR
 #define SPLVAR int splvar
 #undef NETPRI
